import functools
import typing as tp

import jax
import jax.tree_util as jtu
import refx
import refx.tracers
from jax._src import sharding_impls
from refx.partitioning import Partition, Predicate

import nnx
from nnx import scope_lib

A = tp.TypeVar("A")
F = tp.TypeVar("F", bound=tp.Callable[..., tp.Any])
G = tp.TypeVar("G", bound=tp.Callable[..., tp.Any])

AxisName = tp.Hashable
Leaf = tp.Any
Leaves = tp.List[Leaf]


# def dagify(decorator: A, **deco_kwargs) -> A:
#     """Wraps a decorator to make it compatible with refx."""

#     @functools.wraps(decorator)
#     def decorator_wrapper(fun: F) -> F:
#         """"""

#         @functools.wraps(fun)
#         def inner_wrapper(*args, **kwargs) -> tp.Any:
#             trace = refx.tracers.get_top_trace((args, kwargs))
#             with scope_lib.scope(scope_lib.Scope.empty(), trace=trace):
#                 args, kwargs = refx.reref((args, kwargs))
#                 out = fun(*args, **kwargs)
#                 out = refx.deref(out)
#             return out

#         decorated_fun = decorator(inner_wrapper, **deco_kwargs)

#         @functools.wraps(fun)
#         def outer_wrapper(*args_in, **kwargs_in) -> tp.Any:
#             (args, kwargs), dagdef = refx.deref((args_in, kwargs_in))
#             out = decorated_fun(*args, **kwargs)
#             out = refx.reref(out)
#             return out

#         return outer_wrapper

#     return decorator_wrapper


class JitTransform(jax.stages.Wrapped):
    def __init__(
        self,
        fun: tp.Callable[..., tp.Any],
        **jit_kwargs,
    ):
        @functools.partial(jax.jit, **jit_kwargs)
        def jitted_fn(*args, _nnx__dagdef: refx.DagDef, **kwargs):
            top_trace = refx.tracers.current_jax_trace()
            with scope_lib.scope(scope_lib.Scope.empty(), trace=top_trace):
                args, kwargs = refx.reref((args, kwargs), _nnx__dagdef)
                out = fun(*args, **kwargs)
                return refx.deref(out)

        self.jitted_fn = jitted_fn

    def __call__(self, *args, **kwargs):
        (args, kwargs), dagdef = refx.deref((args, kwargs))
        out, dagdef = self.jitted_fn(*args, _nnx__dagdef=dagdef, **kwargs)
        return refx.reref(out, dagdef)

    def __repr__(self):
        return f"JitTransform({self.jitted_fn})"

    def lower(self, *args, **kwargs):
        return self.jitted_fn.lower(*args, **kwargs)


def jit_filter(
    fun: tp.Callable[..., tp.Any],
    *,
    in_shardings: tp.Any = sharding_impls.UNSPECIFIED,
    out_shardings: tp.Any = sharding_impls.UNSPECIFIED,
    static_argnums: tp.Union[int, tp.Sequence[int], None] = None,
    static_argnames: tp.Union[str, tp.Iterable[str], None] = None,
    donate_argnums: tp.Union[int, tp.Sequence[int]] = (),
    keep_unused: bool = False,
    device: tp.Optional[jax.Device] = None,
    backend: tp.Optional[str] = None,
    inline: bool = False,
    abstracted_axes: tp.Optional[tp.Any] = None,
) -> jax.stages.Wrapped:
    """JIT compile a function, dereferencing and rereferencing Refs."""

    if static_argnames is None:
        static_argnames = []
    elif isinstance(static_argnames, str):
        static_argnames = [static_argnames]
    else:
        static_argnames = list(static_argnames)

    static_argnames.append("_nnx__dagdef")

    ref_jit = JitTransform(
        fun,
        in_shardings=in_shardings,
        out_shardings=out_shardings,
        static_argnums=static_argnums,
        static_argnames=static_argnames,
        donate_argnums=donate_argnums,
        keep_unused=keep_unused,
        device=device,
        backend=backend,
        inline=inline,
        abstracted_axes=abstracted_axes,
    )
    ref_jit = functools.wraps(fun)(ref_jit)
    # _update_decorator_fields(ref_jit, fun)
    return ref_jit



import functools
import typing as tp

import jax
import jax.stages
import refx
import refx.tracers
from jax._src import sharding_impls
from refx.partitioning import Partition
import jax.tree_util as jtu

import nnx
from nnx import partitioning, scope_lib

A = tp.TypeVar("A")
F = tp.TypeVar("F", bound=tp.Callable[..., tp.Any])
G = tp.TypeVar("G", bound=tp.Callable[..., tp.Any])

AxisName = tp.Hashable
Leaf = tp.Any
Leaves = tp.List[Leaf]


class JitTransform(jax.stages.Wrapped):
    def __init__(
        self,
        fun: tp.Callable[..., tp.Any],
        stateful: bool,
        **jit_kwargs,
    ):
        @functools.partial(jax.jit, **jit_kwargs)
        def jitted_fn(
            pytree, scope: nnx.Scope, *args, _nnx__dagdef: refx.DagDef, **kwargs
        ):
            top_trace = refx.tracers.current_jax_trace()
            with scope_lib.scope(scope.fork(), trace=top_trace):
                pytree = refx.reref(pytree, _nnx__dagdef)
                out = fun(pytree, *args, **kwargs)
                if self.stateful:
                    out = (refx.deref(pytree), out)
                return out

        self.jitted_fn = jitted_fn
        self.stateful = stateful

    def __call__(self, pytree, *args, **kwargs):
        pytree_in = pytree
        pytree, dagdef = refx.deref(pytree_in)
        scope = nnx.current_scope()
        out = self.jitted_fn(pytree, scope, *args, _nnx__dagdef=dagdef, **kwargs)
        if self.stateful:
            (pytree_out, _), out = out
            refx.update_refs(pytree_in, pytree_out)
        return out

    def __repr__(self):
        return f"JitTransform({self.jitted_fn})"

    def lower(self, *args, **kwargs):
        return self.jitted_fn.lower(*args, **kwargs)


def jit(
    fun: tp.Callable[..., tp.Any],
    *,
    stateful: bool = True,
    in_shardings: tp.Any = sharding_impls.UNSPECIFIED,
    out_shardings: tp.Any = sharding_impls.UNSPECIFIED,
    static_argnums: tp.Union[int, tp.Sequence[int], None] = None,
    static_argnames: tp.Union[str, tp.Iterable[str], None] = None,
    donate_argnums: tp.Union[int, tp.Sequence[int]] = (),
    keep_unused: bool = False,
    device: tp.Optional[jax.Device] = None,
    backend: tp.Optional[str] = None,
    inline: bool = False,
    abstracted_axes: tp.Optional[tp.Any] = None,
) -> jax.stages.Wrapped:
    """JIT compile a function, dereferencing and rereferencing Refs."""

    if static_argnames is None:
        static_argnames = []
    elif isinstance(static_argnames, str):
        static_argnames = [static_argnames]
    else:
        static_argnames = list(static_argnames)

    static_argnames.append("_nnx__dagdef")

    ref_jit = JitTransform(
        fun,
        stateful,
        in_shardings=in_shardings,
        out_shardings=out_shardings,
        static_argnums=static_argnums,
        static_argnames=static_argnames,
        donate_argnums=donate_argnums,
        keep_unused=keep_unused,
        device=device,
        backend=backend,
        inline=inline,
        abstracted_axes=abstracted_axes,
    )
    ref_jit = functools.wraps(fun)(ref_jit)
    # _update_decorator_fields(ref_jit, fun)
    return ref_jit


class GradTransform:
    def __init__(
        self,
        fun: tp.Callable[..., tp.Any],
        stateful: bool,
        predicate: partitioning.Predicate,
        has_aux: bool,
        holomorphic: bool,
        allow_int: bool,
        reduce_axes: tp.Sequence[AxisName],
    ):
        @functools.partial(
            jax.grad,
            argnums=0,  # we'll handle this ourselves
            has_aux=has_aux or stateful,
            holomorphic=holomorphic,
            allow_int=allow_int,
            reduce_axes=reduce_axes,
        )
        def grad_fn(
            diff: Partition,
            non_diff: Partition,
            dagdef: refx.DagDef,
            treedef: jtu.PyTreeDef,
            *args,
        ):
            diff_trace = refx.tracers.get_top_trace(diff)
            scope = scope_lib.current_scope()
            with scope_lib.scope(scope.fork(), trace=diff_trace):
                pytree = refx.merge_partitions((diff, non_diff), treedef)
                pytree = refx.reref(pytree, dagdef)
                out = fun(pytree, *args)

                pytree = refx.deref(pytree)
                if self.has_aux and self.stateful:
                    loss, aux = out
                    out = (loss, (pytree, aux))
                elif self.stateful:
                    out = (out, pytree)

                return out

        self.grad_fn = grad_fn
        self.predicate = predicate
        self.has_aux = has_aux
        self.stateful = stateful

    def __call__(self, pytree, *args):
        pytree_in = pytree
        pytree, dagdef = refx.deref(pytree)
        (diff, nondiff), treedef = refx.tree_partition(pytree, self.predicate)
        # diff, nondiff = refx.deref((diff, nondiff))
        grads = self.grad_fn(diff, nondiff, dagdef, treedef, *args)

        if self.has_aux and self.stateful:
            grad, (pytree_dagdef, aux) = grads
            pytree_out = refx.reref(*pytree_dagdef)
            refx.update_refs(pytree_in, pytree_out)
            return grad, aux
        elif self.stateful:
            grad, pytree_dagdef = grads
            pytree_out = refx.reref(*pytree_dagdef)
            refx.update_refs(pytree_in, pytree_out)
            return grad
        else:
            return grads

    def __repr__(self):
        return f"GradTransform({self.grad_fn})"


@tp.overload
def grad(
    fun: tp.Callable[..., tp.Any],
    wrt: partitioning.CollectionFilter = "params",
    *,
    stateful: bool = True,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., Partition]:
    ...


@tp.overload
def grad(
    fun: tp.Callable[..., tp.Any],
    wrt: partitioning.CollectionFilter = "params",
    *,
    stateful: bool = True,
    has_aux: tp.Literal[True],
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tp.Tuple[Partition, tp.Any]]:
    ...


def grad(
    fun: tp.Callable[..., tp.Any],
    wrt: partitioning.CollectionFilter = "params",
    *,
    stateful: bool = True,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tp.Union[tp.Tuple[Partition, tp.Any], Partition]]:
    predicate = partitioning.to_predicate(wrt)
    ref_grad = GradTransform(
        fun,
        stateful=stateful,
        predicate=predicate,
        has_aux=has_aux,
        holomorphic=holomorphic,
        allow_int=allow_int,
        reduce_axes=reduce_axes,
    )
    ref_grad = functools.wraps(fun)(ref_grad)
    # _update_decorator_fields(ref_grad, fun)
    return ref_grad



from typing import Optional, Sequence

import jax.numpy as jnp
from jax import lax, random

from nnx.nn.module import Module
from nnx import fields, utils


@fields.dataclass
class Dropout(Module):
    """Create a dropout layer.

    Note: When using :meth:`Module.apply() <flax.linen.Module.apply>`, make sure
    to include an RNG seed named `'dropout'`. For example::

      model.apply({'params': params}, inputs=inputs, train=True, rngs={'dropout': dropout_rng})`

    Attributes:
      rate: the dropout probability.  (_not_ the keep rate!)
      broadcast_dims: dimensions that will share the same dropout mask
      deterministic: if false the inputs are scaled by `1 / (1 - rate)` and
        masked, whereas if true, no mask is applied and the inputs are returned
        as is.
      rng_collection: the rng collection name to use when requesting an rng key.
    """

    rate: float
    broadcast_dims: Sequence[int] = ()
    deterministic: Optional[bool] = None
    rng_collection: str = "dropout"

    def __call__(self, inputs, deterministic: Optional[bool] = None):
        """Applies a random dropout mask to the input.

        Args:
          inputs: the inputs that should be randomly masked.
          deterministic: if false the inputs are scaled by `1 / (1 - rate)` and
            masked, whereas if true, no mask is applied and the inputs are returned
            as is.

        Returns:
          The masked inputs reweighted to preserve mean.
        """
        deterministic = utils.first_from(
            deterministic,
            self.deterministic,
            self.get_flag("deterministic", None),
        )

        if (self.rate == 0.0) or deterministic:
            return inputs

        # Prevent gradient NaNs in 1.0 edge-case.
        if self.rate == 1.0:
            return jnp.zeros_like(inputs)

        keep_prob = 1.0 - self.rate
        rng = self.make_rng(self.rng_collection)
        broadcast_shape = list(inputs.shape)
        for dim in self.broadcast_dims:
            broadcast_shape[dim] = 1
        mask = random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
        mask = jnp.broadcast_to(mask, inputs.shape)
        return lax.select(mask, inputs / keep_prob, jnp.zeros_like(inputs))



import typing as tp
import jax
import hashlib

from refx import tracers
import jax.tree_util as jtu


def _stable_hash(data: tp.Tuple[int, ...]) -> int:
    hash_str = " ".join(str(x) for x in data)
    _hash = hashlib.blake2s(hash_str.encode())
    hash_bytes = _hash.digest()
    # uint32 is represented as 4 bytes in big endian
    return int.from_bytes(hash_bytes[:4], byteorder="big")


class RngStream:
    __slots__ = (
        "_key",
        "_count",
        "_count_path",
        "_jax_trace",
        "_refx_trace",
    )

    def __init__(
        self,
        key: jax.random.KeyArray,
        count: int = 0,
        count_path: tp.Tuple[int, ...] = (),
    ):
        self._key = key
        self._count = count
        self._count_path = count_path
        self._jax_trace = tracers.current_jax_trace()
        self._refx_trace = tracers.current_refx_trace()

    def _validate_trace(self):
        if (
            self._jax_trace is not tracers.current_jax_trace()
            or self._refx_trace is not tracers.current_refx_trace()
        ):
            raise ValueError("Rng used in a different trace")

    @property
    def key(self) -> jax.random.KeyArray:
        self._validate_trace()
        return self._key

    @property
    def count(self) -> int:
        return self._count

    @property
    def count_path(self) -> tp.Tuple[int, ...]:
        return self._count_path

    def next(self) -> jax.random.KeyArray:
        self._validate_trace()
        fold_data = _stable_hash(self._count_path + (self._count,))
        self._count += 1
        return jax.random.fold_in(self._key, fold_data)

    def fork(self) -> "RngStream":
        self._validate_trace()
        count_path = self._count_path + (self._count,)
        self._count += 1
        return RngStream(self._key, count_path=count_path)


def _rng_flatten_with_keys(
    rng: RngStream,
) -> tp.Tuple[
    tp.Tuple[tp.Tuple[tp.Hashable, jax.random.KeyArray], ...],
    tp.Tuple[int, tp.Tuple[int, ...]],
]:
    return ((jtu.GetAttrKey("key"), rng.key),), (rng.count, rng.count_path)


def _rng_unflatten(
    aux_data: tp.Tuple[int, tp.Tuple[int, ...]],
    children: tp.Tuple[jax.random.KeyArray, ...],
) -> RngStream:
    count, count_path = aux_data
    key = children[0]
    return RngStream(key, count, count_path)


jax.tree_util.register_pytree_with_keys(
    RngStream, _rng_flatten_with_keys, _rng_unflatten
)



__version__ = "0.0.0"

from refx import Partition

from .fields import dataclass, field, param, ref, static_field
from .filters import jit_filter
from .nn import BatchNorm, Conv, Dropout, Linear, Module, ModuleDef
from .nn.activations import (
    celu,
    elu,
    gelu,
    glu,
    hard_sigmoid,
    hard_silu,
    hard_swish,
    hard_tanh,
    leaky_relu,
    log_sigmoid,
    log_softmax,
    logsumexp,
    normalize,
    one_hot,
    relu,
    relu6,
    selu,
    sigmoid,
    silu,
    soft_sign,
    softmax,
    softplus,
    standardize,
    swish,
    tanh,
)
from .rng_stream import RngStream
from .scope_lib import (
    Scope,
    apply,
    current_scope,
    fork_scope,
    get_flag,
    init,
    make_rng,
    scope,
)
from .transforms import grad, jit



import typing as tp

import jax
import jax.numpy as jnp
from jax import lax

from nnx.nn.module import Module
from nnx.nn import initializers, dtypes
from nnx import fields, utils

PRNGKey = jax.Array
Array = jax.Array
Shape = tp.Tuple[int, ...]
Dtype = tp.Any  # this could be a real type?

Axes = tp.Union[int, tp.Any]


def _canonicalize_axes(rank: int, axes: Axes) -> tp.Tuple[int, ...]:
    """Returns a tuple of deduplicated, sorted, and positive axes."""
    if not isinstance(axes, tp.Iterable):
        axes = (axes,)
    return tuple(set([rank + axis if axis < 0 else axis for axis in axes]))


def _abs_sq(x):
    """Computes the elementwise square of the absolute value |x|^2."""
    if jnp.iscomplexobj(x):
        return lax.square(lax.real(x)) + lax.square(lax.imag(x))
    else:
        return lax.square(x)


def _compute_stats(
    x: Array,
    axes: tp.Optional[Axes],
    dtype: tp.Optional[Dtype],
    axis_name: tp.Optional[str] = None,
    axis_index_groups: tp.Any = None,
    use_mean: bool = True,
):
    """Computes mean and variance statistics.

    This implementation takes care of a few important details:
    - Computes in float32 precision for stability in half precision training.
    - mean and variance are computable in a single XLA fusion,
      by using Var = E[|x|^2] - |E[x]|^2 instead of Var = E[|x - E[x]|^2]).
    - Clips negative variances to zero which can happen due to
      roundoff errors. This avoids downstream NaNs.
    - Supports averaging across a parallel axis and subgroups of a parallel axis
      with a single `lax.pmean` call to avoid latency.

    Arguments:
      x: Input array.
      axes: The axes in ``x`` to compute mean and variance statistics for.
      dtype: tp.Optional dtype specifying the minimal precision. Statistics
        are always at least float32 for stability (default: dtype of x).
      axis_name: tp.Optional name for the pmapped axis to compute mean over.
      axis_index_groups: tp.Optional axis indices.
      use_mean: If true, calculate the mean from the input and use it when
        computing the variance. If false, set the mean to zero and compute
        the variance without subtracting the mean.

    Returns:
      A pair ``(mean, var)``.
    """
    if dtype is None:
        dtype = jnp.result_type(x)
    # promote x to at least float32, this avoids half precision computation
    # but preserves double or complex floating points
    dtype = jnp.promote_types(dtype, jnp.float32)
    x = jnp.asarray(x, dtype)

    mean2 = jnp.mean(_abs_sq(x), axes)
    if use_mean:
        mean = jnp.mean(x, axes)
    else:
        mean = jnp.zeros(mean2.shape, dtype=dtype)

    if axis_name is not None:
        concatenated_mean = jnp.concatenate([mean, mean2])
        mean, mean2 = jnp.split(
            lax.pmean(
                concatenated_mean,
                axis_name=axis_name,
                axis_index_groups=axis_index_groups,
            ),
            2,
        )
    # mean2 - _abs_sq(mean) is not guaranteed to be non-negative due
    # to floating point round-off errors.
    var = jnp.maximum(0.0, mean2 - _abs_sq(mean))
    return mean, var


def _normalize(
    x: Array,
    mean: Array,
    var: Array,
    scale: tp.Optional[Array],
    bias: tp.Optional[Array],
    reduction_axes: Axes,
    feature_axes: Axes,
    dtype: Dtype,
    epsilon: float,
):
    """ "Normalizes the input of a normalization layer and optionally applies a learned scale and bias.

    Arguments:
      x: The input.
      mean: Mean to use for normalization.
      var: Variance to use for normalization.
      reduction_axes: The axes in ``x`` to reduce.
      feature_axes: Axes containing features. A separate bias and scale is learned
        for each specified feature.
      dtype: The dtype of the result (default: infer from input and params).
      epsilon: Normalization epsilon.

    Returns:
      The normalized input.
    """
    reduction_axes = _canonicalize_axes(x.ndim, reduction_axes)
    feature_axes = _canonicalize_axes(x.ndim, feature_axes)
    stats_shape = list(x.shape)
    for axis in reduction_axes:
        stats_shape[axis] = 1
    mean = mean.reshape(stats_shape)
    var = var.reshape(stats_shape)
    feature_shape = [1] * x.ndim
    reduced_feature_shape = []
    for ax in feature_axes:
        feature_shape[ax] = x.shape[ax]
        reduced_feature_shape.append(x.shape[ax])
    y = x - mean
    mul = lax.rsqrt(var + epsilon)
    args = [x]
    if scale is not None:
        scale = scale.reshape(feature_shape)
        mul *= scale
        args.append(scale)
    y *= mul
    if bias is not None:
        bias = bias.reshape(feature_shape)
        y += bias
        args.append(bias)
    dtype = dtypes.canonicalize_dtype(*args, dtype=dtype)
    return jnp.asarray(y, dtype)


@fields.dataclass
class BatchNorm(Module):
    """BatchNorm Module.

    Usage Note:
    If we define a model with BatchNorm, for example::

      BN = nn.BatchNorm(use_running_average=False, momentum=0.9, epsilon=1e-5,
                        dtype=jnp.float32)

    The initialized variables dict will contain in addition to a 'params'
    collection a separate 'batch_stats' collection that will contain all the
    running statistics for all the BatchNorm layers in a model::

      vars_initialized = BN.init(key, x)  # {'params': ..., 'batch_stats': ...}

    We then update the batch_stats during training by specifying that the
    `batch_stats` collection is mutable in the `apply` method for our module.::

      vars_in = {'params': params, 'batch_stats': old_batch_stats}
      y, mutated_vars = BN.apply(vars_in, x, mutable=['batch_stats'])
      new_batch_stats = mutated_vars['batch_stats']

    During eval we would define BN with `use_running_average=True` and use the
    batch_stats collection from training to set the statistics.  In this case
    we are not mutating the batch statistics collection, and needn't mark it
    mutable::

      vars_in = {'params': params, 'batch_stats': training_batch_stats}
      y = BN.apply(vars_in, x)

    Attributes:
      use_running_average: if True, the statistics stored in batch_stats
        will be used instead of computing the batch statistics on the input.
      axis: the feature or non-batch axis of the input.
      momentum: decay rate for the exponential moving average of
        the batch statistics.
      epsilon: a small float added to variance to avoid dividing by zero.
      dtype: the dtype of the result (default: infer from input and params).
      param_dtype: the dtype passed to parameter initializers (default: float32).
      use_bias:  if True, bias (beta) is added.
      use_scale: if True, multiply by scale (gamma).
        When the next layer is linear (also e.g. nn.relu), this can be disabled
        since the scaling will be done by the next layer.
      bias_init: initializer for bias, by default, zero.
      scale_init: initializer for scale, by default, one.
      axis_name: the axis name used to combine batch statistics from multiple
        devices. See `jax.pmap` for a description of axis names (default: None).
      axis_index_groups: groups of axis indices within that named axis
        representing subsets of devices to reduce over (default: None). For
        example, `[[0, 1], [2, 3]]` would independently batch-normalize over
        the examples on the first two and last two devices. See `jax.lax.psum`
        for more details.
    """

    num_features: int = fields.static_field()
    use_running_average: tp.Optional[bool] = fields.static_field(default=None)
    axis: int = fields.static_field(default=-1)
    momentum: float = fields.static_field(default=0.99)
    epsilon: float = fields.static_field(default=1e-5)
    dtype: tp.Optional[Dtype] = fields.static_field(default=None)
    param_dtype: Dtype = fields.static_field(default=jnp.float32)
    use_bias: bool = fields.static_field(default=True)
    use_scale: bool = fields.static_field(default=True)
    bias_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = fields.static_field(
        default=initializers.zeros()
    )
    scale_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = fields.static_field(
        default=initializers.ones()
    )
    axis_name: tp.Optional[str] = fields.static_field(default=None)
    axis_index_groups: tp.Any = fields.static_field(default=None)
    # refs
    mean: Array = fields.ref("batch_stats", init=False)
    var: Array = fields.ref("batch_stats", init=False)
    scale: tp.Optional[Array] = fields.param(init=False)
    bias: tp.Optional[Array] = fields.param(init=False)

    def __post_init__(self):
        feature_shape = (self.num_features,)
        self.mean = jnp.zeros(feature_shape, jnp.float32)
        self.var = jnp.ones(feature_shape, jnp.float32)

        if self.use_scale:
            key = self.make_rng("params")
            self.scale = self.scale_init(key, feature_shape, self.param_dtype)
        else:
            self.scale = None

        if self.use_bias:
            key = self.make_rng("params")
            self.bias = self.bias_init(key, feature_shape, self.param_dtype)
        else:
            self.bias = None

    def __call__(self, x, use_running_average: tp.Optional[bool] = None):
        """Normalizes the input using batch statistics.

        Args:
          x: the input to be normalized.
          use_running_average: if true, the statistics stored in batch_stats
            will be used instead of computing the batch statistics on the input.

        Returns:
          Normalized inputs (the same shape as inputs).
        """

        use_running_average = utils.first_from(
            use_running_average,
            self.use_running_average,
            self.get_flag("use_running_average", None),
        )
        feature_axes = _canonicalize_axes(x.ndim, self.axis)
        reduction_axes = tuple(i for i in range(x.ndim) if i not in feature_axes)

        if use_running_average:
            mean, var = self.mean, self.var
        else:
            mean, var = _compute_stats(
                x,
                reduction_axes,
                dtype=self.dtype,
                axis_name=self.axis_name,
                axis_index_groups=self.axis_index_groups,
            )

            self.mean = self.momentum * self.mean + (1 - self.momentum) * mean
            self.var = self.momentum * self.var + (1 - self.momentum) * var

        return _normalize(
            x,
            mean,
            var,
            self.scale,
            self.bias,
            reduction_axes,
            feature_axes,
            self.dtype,
            self.epsilon,
        )



from jax.nn import celu
from jax.nn import elu
from jax.nn import gelu
from jax.nn import glu
from jax.nn import hard_sigmoid
from jax.nn import hard_silu
from jax.nn import hard_swish
from jax.nn import hard_tanh
from jax.nn import leaky_relu
from jax.nn import log_sigmoid
from jax.nn import log_softmax
from jax.nn import logsumexp
from jax.nn import normalize
from jax.nn import one_hot
from jax.nn import relu
from jax.nn import relu6
from jax.nn import selu
from jax.nn import sigmoid
from jax.nn import silu
from jax.nn import soft_sign
from jax.nn import softmax
from jax.nn import softplus
from jax.nn import standardize
from jax.nn import swish
from jax.numpy import tanh

__all__ = [
    "celu",
    "elu",
    "gelu",
    "glu",
    "hard_sigmoid",
    "hard_silu",
    "hard_swish",
    "hard_tanh",
    "leaky_relu",
    "log_sigmoid",
    "log_softmax",
    "logsumexp",
    "normalize",
    "one_hot",
    "relu",
    "relu6",
    "selu",
    "sigmoid",
    "silu",
    "soft_sign",
    "softmax",
    "softplus",
    "standardize",
    "swish",
    "tanh",
]



import typing as tp

import jax
from jax import lax
import jax.numpy as jnp
import numpy as np

from nnx.nn.module import Module
from nnx.nn import initializers
from nnx import fields
from nnx.nn import dtypes

Array = jax.Array
PRNGKey = tp.Any
Shape = tp.Tuple[int, ...]
Dtype = tp.Any  # this could be a real type?
PrecisionLike = tp.Union[
    None, str, lax.Precision, tp.Tuple[str, str], tp.Tuple[lax.Precision, lax.Precision]
]
ConvGeneralDilatedT = tp.Callable[..., Array]
PaddingLike = tp.Union[str, int, tp.Sequence[tp.Union[int, tp.Tuple[int, int]]]]
LaxPadding = tp.Union[str, tp.Sequence[tp.Tuple[int, int]]]
DotGeneralT = tp.Callable[..., Array]


default_kernel_init = initializers.lecun_normal()


def canonicalize_padding(padding: PaddingLike, rank: int) -> LaxPadding:
    """ "Canonicalizes conv padding to a jax.lax supported format."""
    if isinstance(padding, str):
        return padding
    if isinstance(padding, int):
        return [(padding, padding)] * rank
    if isinstance(padding, tp.Sequence) and len(padding) == rank:
        new_pad = []
        for p in padding:
            if isinstance(p, int):
                new_pad.append((p, p))
            elif isinstance(p, tuple) and len(p) == 2:
                new_pad.append(p)
            else:
                break
        if len(new_pad) == rank:
            return new_pad
    raise ValueError(
        f"Invalid padding format: {padding}, should be str, int,"
        f" or a sequence of len {rank} where each element is an"
        f" int or pair of ints."
    )


def _conv_dimension_numbers(input_shape):
    """Computes the dimension numbers based on the input shape."""
    ndim = len(input_shape)
    lhs_spec = (0, ndim - 1) + tuple(range(1, ndim - 1))
    rhs_spec = (ndim - 1, ndim - 2) + tuple(range(0, ndim - 2))
    out_spec = lhs_spec
    return lax.ConvDimensionNumbers(lhs_spec, rhs_spec, out_spec)


@fields.dataclass
class Linear(Module):
    """A linear transformation applied over the last dimension of the input.

    Attributes:
      features: the number of output features.
      use_bias: whether to add a bias to the output (default: True).
      dtype: the dtype of the computation (default: infer from input and params).
      param_dtype: the dtype passed to parameter initializers (default: float32).
      precision: numerical precision of the computation see `jax.lax.Precision`
        for details.
      kernel_init: initializer function for the weight matrix.
      bias_init: initializer function for the bias.
    """

    in_features: int = fields.static_field()
    out_features: int = fields.static_field()
    use_bias: bool = fields.static_field(default=True)
    dtype: tp.Optional[Dtype] = fields.static_field(default=None)
    param_dtype: Dtype = fields.static_field(default=jnp.float32)
    precision: PrecisionLike = fields.static_field(default=None)
    kernel_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = fields.static_field(
        default=default_kernel_init
    )
    bias_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = fields.static_field(
        default=initializers.zeros()
    )
    dot_general: DotGeneralT = fields.static_field(default=lax.dot_general)
    # ref fields
    kernel: Array = fields.param(init=False)
    bias: tp.Optional[Array] = fields.param(init=False)

    def __post_init__(self):
        kernel_key = self.make_rng("params")
        self.kernel = self.kernel_init(
            kernel_key, (self.in_features, self.out_features), self.param_dtype
        )
        if self.use_bias:
            bias_key = self.make_rng("params")
            self.bias = self.bias_init(bias_key, (self.out_features,), self.param_dtype)
        else:
            self.bias = None

    def __call__(self, inputs: Array) -> Array:
        """Applies a linear transformation to the inputs along the last dimension.

        Args:
          inputs: The nd-array to be transformed.

        Returns:
          The transformed input.
        """
        kernel = self.kernel
        bias = self.bias

        inputs, kernel, bias = dtypes.promote_dtype(
            inputs, kernel, bias, dtype=self.dtype
        )
        y = self.dot_general(
            inputs,
            kernel,
            (((inputs.ndim - 1,), (0,)), ((), ())),
            precision=self.precision,
        )
        if bias is not None:
            y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
        return y


@fields.dataclass
class Conv(Module):
    """Convolution Module wrapping `lax.conv_general_dilated[_local]`.

    Attributes:
      features: number of convolution filters.
      kernel_size: shape of the convolutional kernel. For 1D convolution,
        the kernel size can be passed as an integer. For all other cases, it must
        be a sequence of integers.
      strides: an integer or a sequence of `n` integers, representing the
        inter-window strides (default: 1).
      padding: either the string `'SAME'`, the string `'VALID'`, the string
        `'CIRCULAR'` (periodic boundary conditions), or a sequence of `n` `(low,
        high)` integer pairs that give the padding to apply before and after each
        spatial dimension. A single int is interpeted as applying the same padding
        in all dims and passign a single int in a sequence causes the same padding
        to be used on both sides. `'CAUSAL'` padding for a 1D convolution will
        left-pad the convolution axis, resulting in same-sized output.
      input_dilation: an integer or a sequence of `n` integers, giving the
        dilation factor to apply in each spatial dimension of `inputs`
        (default: 1). Convolution with input dilation `d` is equivalent to
        transposed convolution with stride `d`.
      kernel_dilation: an integer or a sequence of `n` integers, giving the
        dilation factor to apply in each spatial dimension of the convolution
        kernel (default: 1). Convolution with kernel dilation
        is also known as 'atrous convolution'.
      feature_group_count: integer, default 1. If specified divides the input
        features into groups.
      use_bias: whether to add a bias to the output (default: True).
      mask: Optional mask for the weights during masked convolution. The mask must
            be the same shape as the convolution weight matrix.
      dtype: the dtype of the computation (default: infer from input and params).
      param_dtype: the dtype passed to parameter initializers (default: float32).
      precision: numerical precision of the computation see `jax.lax.Precision`
        for details.
      kernel_init: initializer for the convolutional kernel.
      bias_init: initializer for the bias.
    """

    in_features: int = fields.static_field()
    out_features: int = fields.static_field()
    kernel_size: tp.Sequence[int] = fields.static_field()
    strides: tp.Union[None, int, tp.Sequence[int]] = fields.static_field(default=1)
    padding: PaddingLike = fields.static_field(default="SAME")
    input_dilation: tp.Union[None, int, tp.Sequence[int]] = fields.static_field(
        default=1
    )
    kernel_dilation: tp.Union[None, int, tp.Sequence[int]] = fields.static_field(
        default=1
    )
    feature_group_count: int = fields.static_field(default=1)
    use_bias: bool = fields.static_field(default=True)
    mask_fn: tp.Optional[tp.Callable[[Array], Array]] = fields.static_field(
        default=None
    )
    dtype: tp.Optional[Dtype] = fields.static_field(default=None)
    param_dtype: Dtype = fields.static_field(default=jnp.float32)
    precision: PrecisionLike = fields.static_field(default=None)
    kernel_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = fields.static_field(
        default=default_kernel_init
    )
    bias_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = fields.static_field(
        default=initializers.zeros()
    )
    conv_general_dilated: ConvGeneralDilatedT = fields.static_field(
        default=lax.conv_general_dilated
    )
    # ref fields
    kernel: Array = fields.param(init=False)
    bias: tp.Optional[Array] = fields.param(init=False)

    def __post_init__(self):
        if isinstance(self.kernel_size, int):
            raise TypeError(
                "Expected Conv kernel_size to be a"
                " tuple/list of integers (eg.: [3, 3]) but got"
                f" {self.kernel_size}."
            )
        else:
            self.kernel_size = tuple(self.kernel_size)

        kernel_shape = self.kernel_size + (
            self.in_features // self.feature_group_count,
            self.out_features,
        )
        kernel_key = self.make_rng("params")
        self.kernel = self.kernel_init(kernel_key, kernel_shape, self.param_dtype)

        if self.use_bias:
            bias_shape = (self.out_features,)
            bias_key = self.make_rng("params")
            self.bias = self.bias_init(bias_key, bias_shape, self.param_dtype)
        else:
            self.bias = None

    def __call__(self, inputs: Array) -> Array:
        """Applies a (potentially unshared) convolution to the inputs.

        Args:
          inputs: input data with dimensions (*batch_dims, spatial_dims...,
            features). This is the channels-last convention, i.e. NHWC for a 2d
            convolution and NDHWC for a 3D convolution. Note: this is different from
            the input convention used by `lax.conv_general_dilated`, which puts the
            spatial dimensions last.
            Note: If the input has more than 1 batch dimension, all batch dimensions
            are flattened into a single dimension for the convolution and restored
            before returning.  In some cases directly vmap'ing the layer may yield
            better performance than this default flattening approach.  If the input
            lacks a batch dimension it will be added for the convolution and removed
            n return, an allowance made to enable writing single-example code.

        Returns:
          The convolved data.
        """

        assert isinstance(self.kernel_size, tuple)
        kernel_size = self.kernel_size

        def maybe_broadcast(
            x: tp.Optional[tp.Union[int, tp.Sequence[int]]]
        ) -> tp.Tuple[int, ...]:
            if x is None:
                # backward compatibility with using None as sentinel for
                # broadcast 1
                x = 1
            if isinstance(x, int):
                return (x,) * len(kernel_size)
            return tuple(x)

        # Combine all input batch dimensions into a single leading batch axis.
        num_batch_dimensions = inputs.ndim - (len(kernel_size) + 1)
        if num_batch_dimensions != 1:
            input_batch_shape = inputs.shape[:num_batch_dimensions]
            total_batch_size = int(np.prod(input_batch_shape))
            flat_input_shape = (total_batch_size,) + inputs.shape[num_batch_dimensions:]
            inputs = jnp.reshape(inputs, flat_input_shape)

        # self.strides or (1,) * (inputs.ndim - 2)
        strides = maybe_broadcast(self.strides)
        input_dilation = maybe_broadcast(self.input_dilation)
        kernel_dilation = maybe_broadcast(self.kernel_dilation)

        padding_lax = canonicalize_padding(self.padding, len(kernel_size))
        if padding_lax == "CIRCULAR":
            kernel_size_dilated = [
                (k - 1) * d + 1 for k, d in zip(kernel_size, kernel_dilation)
            ]
            zero_pad: tp.List[tp.Tuple[int, int]] = [(0, 0)]
            pads = (
                zero_pad
                + [((k - 1) // 2, k // 2) for k in kernel_size_dilated]
                + [(0, 0)]
            )
            inputs = jnp.pad(inputs, pads, mode="wrap")
            padding_lax = "VALID"
        elif padding_lax == "CAUSAL":
            if len(kernel_size) != 1:
                raise ValueError(
                    "Causal padding is only implemented for 1D convolutions."
                )
            left_pad = kernel_dilation[0] * (kernel_size[0] - 1)
            pads = [(0, 0), (left_pad, 0), (0, 0)]
            inputs = jnp.pad(inputs, pads)
            padding_lax = "VALID"

        dimension_numbers = _conv_dimension_numbers(inputs.shape)

        # One shared convolutional kernel for all pixels in the output.
        assert self.in_features % self.feature_group_count == 0

        kernel = self.kernel

        if self.mask_fn is not None:
            kernel = self.mask_fn(kernel)

        bias = self.bias

        inputs, kernel, bias = dtypes.promote_dtype(
            inputs, kernel, bias, dtype=self.dtype
        )

        y = self.conv_general_dilated(
            inputs,
            kernel,
            strides,
            padding_lax,
            lhs_dilation=input_dilation,
            rhs_dilation=kernel_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if self.use_bias:
            bias = bias.reshape((1,) * (y.ndim - bias.ndim) + bias.shape)
            y += bias

        if num_batch_dimensions != 1:
            output_shape = input_batch_shape + y.shape[1:]
            y = jnp.reshape(y, output_shape)
        return y



from typing import Any, Optional, List

from jax import numpy as jnp
import jax


Dtype = Any
Array = Any


def canonicalize_dtype(
    *args, dtype: Optional[Dtype] = None, inexact: bool = True
) -> Dtype:
    """Canonicalize an optional dtype to the definitive dtype.

    If the ``dtype`` is None this function will infer the dtype. If it is not
    None it will be returned unmodified or an exceptions is raised if the dtype
    is invalid.
    from the input arguments using ``jnp.result_type``.

    Args:
      *args: JAX array compatible values. None values
        are ignored.
      dtype: Optional dtype override. If specified the arguments are cast to
        the specified dtype instead and dtype inference is disabled.
      inexact: When True, the output dtype must be a subdtype
      of `jnp.inexact`. Inexact dtypes are real or complex floating points. This
      is useful when you want to apply operations that don't work directly on
      integers like taking a mean for example.
    Returns:
      The dtype that *args should be cast to.
    """
    if dtype is None:
        args_filtered = [jnp.asarray(x) for x in args if x is not None]
        dtype = jnp.result_type(*args_filtered)
        if inexact and not jnp.issubdtype(dtype, jnp.inexact):
            dtype = jnp.promote_types(jnp.float32, dtype)
    if inexact and not jnp.issubdtype(dtype, jnp.inexact):
        raise ValueError(f"Dtype must be inexact: {dtype}")
    return dtype


def promote_dtype(*args, dtype=None, inexact=True) -> List[Array]:
    """ "Promotes input arguments to a specified or inferred dtype.

    All args are cast to the same dtype. See ``canonicalize_dtype`` for how
    this dtype is determined.

    The behavior of promote_dtype is mostly a convinience wrapper around
    ``jax.numpy.promote_types``. The differences being that it automatically casts
    all input to the inferred dtypes, allows inference to be overridden by a
    forced dtype, and has an optional check to garantuee the resulting dtype is
    inexact.

    Args:
      *args: JAX array compatible values. None values
        are returned as is.
      dtype: Optional dtype override. If specified the arguments are cast to
        the specified dtype instead and dtype inference is disabled.
      inexact: When True, the output dtype must be a subdtype
      of `jnp.inexact`. Inexact dtypes are real or complex floating points. This
      is useful when you want to apply operations that don't work directly on
      integers like taking a mean for example.
    Returns:
      The arguments cast to arrays of the same dtype.
    """
    dtype = canonicalize_dtype(*args, dtype=dtype, inexact=inexact)
    return [jnp.asarray(x, dtype) if x is not None else None for x in args]



import contextlib
import typing as tp

A = tp.TypeVar("A")


@contextlib.contextmanager
def contexts(*contexts: tp.ContextManager[tp.Any]) -> tp.Iterator[tp.Tuple[tp.Any]]:
    """Context manager that enters multiple contexts."""
    with contextlib.ExitStack() as stack:
        yield tuple(stack.enter_context(context) for context in contexts)


def first_from(*args: tp.Optional[A]) -> A:
    """Return the first non-None argument."""
    for arg in args:
        if arg is not None:
            return arg
    raise ValueError("No non-None arguments found.")



import dataclasses
import functools
import typing as tp
import jax

import refx

Predicate = tp.Callable[[tp.Any], bool]
CollectionPredicate = tp.Callable[[tp.Hashable], bool]
CollectionFilter = tp.Union[
    str, tp.Sequence[str], CollectionPredicate, tp.Sequence[CollectionPredicate]
]
LeafPredicate = tp.Callable[[tp.Any], bool]


def tree_partition(
    pytree,
    *filters: CollectionFilter,
    is_leaf: tp.Optional[LeafPredicate] = None,
) -> tp.Tuple[tp.Tuple[refx.Partition, ...], jax.tree_util.PyTreeDef]:
    predicates = (to_predicate(filter) for filter in filters)
    return refx.tree_partition(pytree, *predicates, is_leaf=is_leaf)


def get_partition(
    pytree,
    filter: CollectionFilter,
    is_leaf: tp.Optional[LeafPredicate] = None,
) -> refx.Partition:
    predicate = to_predicate(filter)
    return refx.get_partition(pytree, predicate, is_leaf=is_leaf)


def filter_collection(
    f: tp.Callable[[CollectionFilter], CollectionPredicate]
) -> tp.Callable[[CollectionFilter], Predicate]:
    @functools.wraps(f)
    def wrapper(collection_filter: CollectionFilter) -> Predicate:
        collection_predicate = f(collection_filter)
        return lambda x: isinstance(x, refx.Referential) and collection_predicate(
            x.collection
        )

    return wrapper


@filter_collection
def to_predicate(collection_filter: CollectionFilter) -> CollectionPredicate:
    if isinstance(collection_filter, str):
        return Is(collection_filter)
    elif isinstance(collection_filter, tp.Sequence):
        return Any(collection_filter)
    elif callable(collection_filter):
        return collection_filter
    else:
        raise TypeError(f"Invalid collection filter: {collection_filter}")


merge_partitions = refx.merge_partitions


@dataclasses.dataclass
class Is:
    collection: str

    def __call__(self, collection):
        return collection == self.collection


class Any:
    def __init__(self, collection_filters: tp.Sequence[CollectionFilter]):
        self.predicates = tuple(
            to_predicate(collection_filter) for collection_filter in collection_filters
        )

    def __call__(self, collection):
        return any(predicate(collection) for predicate in self.predicates)


class Not:
    def __init__(self, collection_filter: CollectionFilter):
        self.predicate = to_predicate(collection_filter)

    def __call__(self, x):
        return not self.predicate(x)



import jax
from jax.nn.initializers import constant as constant
from jax.nn.initializers import delta_orthogonal as delta_orthogonal
from jax.nn.initializers import glorot_normal as glorot_normal
from jax.nn.initializers import glorot_uniform as glorot_uniform
from jax.nn.initializers import he_normal as he_normal
from jax.nn.initializers import he_uniform as he_uniform
from jax.nn.initializers import kaiming_normal as kaiming_normal
from jax.nn.initializers import kaiming_uniform as kaiming_uniform
from jax.nn.initializers import lecun_normal as lecun_normal
from jax.nn.initializers import lecun_uniform as lecun_uniform
from jax.nn.initializers import normal as normal
from jax.nn.initializers import orthogonal as orthogonal
from jax.nn.initializers import uniform as uniform
from jax.nn.initializers import variance_scaling as variance_scaling
from jax.nn.initializers import xavier_normal as xavier_normal
from jax.nn.initializers import xavier_uniform as xavier_uniform
from jax.nn.initializers import Initializer as Initializer


def zeros() -> Initializer:
    """Builds an initializer that returns a constant array full of zeros.

    >>> import jax, jax.numpy as jnp
    >>> from flax.linen.initializers import zeros_init
    >>> zeros_initializer = zeros_init()
    >>> zeros_initializer(jax.random.PRNGKey(42), (2, 3), jnp.float32)
    Array([[0., 0., 0.],
           [0., 0., 0.]], dtype=float32)
    """
    return jax.nn.initializers.zeros


def ones() -> Initializer:
    """Builds an initializer that returns a constant array full of ones.

    >>> import jax, jax.numpy as jnp
    >>> from flax.linen.initializers import ones_init
    >>> ones_initializer = ones_init()
    >>> ones_initializer(jax.random.PRNGKey(42), (3, 2), jnp.float32)
    Array([[1., 1.],
           [1., 1.],
           [1., 1.]], dtype=float32)
    """
    return jax.nn.initializers.ones



import contextlib
import dataclasses
import threading
import typing as tp
from types import MappingProxyType

import jax
import jax.tree_util as jtu
import refx
from refx import tracers

from nnx import utils
from nnx.rng_stream import RngStream

KeyArray = jax.random.KeyArray


class Scope:
    __slots__ = ("_rng_streams", "_flags")

    def __init__(
        self,
        rng_streams: tp.Mapping[str, RngStream],
        flags: tp.Mapping[str, tp.Hashable],
    ):
        self._rng_streams = MappingProxyType(rng_streams)
        self._flags = MappingProxyType(flags)

    @classmethod
    def empty(cls) -> "Scope":
        return Scope({}, {})

    @classmethod
    def from_keys_and_flags(
        cls, keys: tp.Mapping[str, KeyArray], flags: tp.Mapping[str, tp.Hashable]
    ) -> "Scope":
        rng_streams = {k: RngStream(v) for k, v in keys.items()}
        return Scope(rng_streams, flags)

    @property
    def rng_streams(self) -> tp.Mapping[str, RngStream]:
        return self._rng_streams

    @property
    def flags(self) -> tp.Mapping[str, tp.Hashable]:
        return self._flags

    def fork(self) -> "Scope":
        rng_streams = {k: v.fork() for k, v in self._rng_streams.items()}
        return Scope(rng_streams, self._flags)

    def unsafe_trace_update(self):
        for rng_stream in self._rng_streams.values():
            rng_stream._jax_trace = tracers.current_jax_trace()
            rng_stream._refx_trace = tracers.current_refx_trace()

    def copy(self) -> "Scope":
        return Scope(self._rng_streams, self._flags)


def _scope_flatten_with_keys(scope: Scope):
    return ((jtu.GetAttrKey("rng_streams"), scope._rng_streams.copy()),), scope._flags


def _scope_unflatten(flags, children):
    return Scope(children[0], flags)


jtu.register_pytree_with_keys(Scope, _scope_flatten_with_keys, _scope_unflatten)


@dataclasses.dataclass
class _ScopeContext(threading.local):
    scope_stack: tp.List[Scope] = dataclasses.field(
        default_factory=lambda: [Scope.empty()]
    )


_CONTEXT = _ScopeContext()


def current_scope() -> Scope:
    return _CONTEXT.scope_stack[-1]


@tp.overload
def scope(
    rngs: tp.Optional[tp.Mapping[tp.Hashable, KeyArray]] = None,
    /,
    *,
    flags: tp.Optional[tp.Mapping[str, tp.Hashable]] = None,
    trace: tp.Optional[tracers.MainTrace] = None,
    mutable: tp.Optional[tp.Callable[[tp.Hashable], bool]] = None,
) -> tp.ContextManager[None]:
    ...


@tp.overload
def scope(
    scope: Scope,
    /,
    *,
    trace: tp.Optional[tracers.MainTrace] = None,
    mutable: tp.Optional[tp.Callable[[tp.Hashable], bool]] = None,
) -> tp.ContextManager[None]:
    ...


@contextlib.contextmanager
def scope(
    rngs_or_scope: tp.Union[tp.Mapping[str, KeyArray], Scope, None] = None,
    /,
    *,
    flags: tp.Optional[tp.Mapping[str, tp.Hashable]] = None,
    trace: tp.Optional[tracers.MainTrace] = None,
    mutable: tp.Optional[tp.Callable[[tp.Hashable], bool]] = None,
):
    if isinstance(rngs_or_scope, Scope):
        if flags is not None:
            raise ValueError("Cannot set flags when passing a Scope")
        scope = rngs_or_scope.copy()
    else:
        if flags is None:
            flags = current_scope().flags
        if rngs_or_scope is None:
            rngs_or_scope = current_scope().rng_streams

        scope = Scope.from_keys_and_flags(rngs_or_scope, flags)
    _CONTEXT.scope_stack.append(scope)

    _contexts = []

    if trace is not None:
        _contexts.append(tracers.refx_trace(trace))

    if mutable is not None:
        _contexts.append(refx.mutable(mutable))

    try:
        with utils.contexts(*_contexts):
            scope.unsafe_trace_update()
            yield
    finally:
        _CONTEXT.scope_stack.pop()


def fork_scope():
    return scope(current_scope().fork())


def make_rng(collection: tp.Hashable) -> KeyArray:
    scope = current_scope()
    if collection not in scope.rng_streams:
        raise ValueError(f"Unknown collection: {collection}")
    return scope.rng_streams[collection].next()


def get_flag(name: str, default: tp.Any = dataclasses.MISSING) -> tp.Hashable:
    scope = current_scope()
    if name not in scope.flags:
        if default is dataclasses.MISSING:
            raise ValueError(f"Unknown flag: {name}")
        return default
    return scope.flags[name]


def all_mutable(_):
    return True


def init(
    rngs: tp.Union[KeyArray, tp.Dict[tp.Hashable, KeyArray], None] = None,
    /,
    *,
    mutable: tp.Callable[[tp.Hashable], bool] = all_mutable,
    flags: tp.Optional[tp.Dict[str, tp.Hashable]] = None,
    trace: tp.Optional[tracers.MainTrace] = None,
) -> tp.ContextManager[None]:
    if rngs is None:
        rngs = {}
    elif isinstance(rngs, jax.Array):
        rngs = {"params": rngs}

    if flags is None:
        flags = {}

    flags["initializing"] = True

    return scope(rngs, flags=flags, trace=trace, mutable=mutable)


def apply(
    rngs: tp.Union[KeyArray, tp.Dict[tp.Hashable, KeyArray], None] = None,
    /,
    *,
    mutable: tp.Callable[[tp.Hashable], bool] = all_mutable,
    flags: tp.Optional[tp.Dict[str, tp.Hashable]] = None,
    trace: tp.Optional[tracers.MainTrace] = None,
) -> tp.ContextManager[None]:
    if rngs is None:
        rngs = {}
    elif isinstance(rngs, jax.Array):
        rngs = {"dropout": rngs}

    if flags is None:
        flags = {}

    flags["initializing"] = False

    return scope(rngs, flags=flags, trace=trace, mutable=mutable)



import dataclasses
from dataclasses import field
from simple_pytree import static_field
import refx
import typing as tp
import typing_extensions as tpe

A = tp.TypeVar("A")


# ----------------------------------------
# fields
# ----------------------------------------


def ref(
    collection: str,
    default: tp.Any = dataclasses.MISSING,
    *,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
) -> tp.Any:
    return refx.RefField(
        collection=collection,
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def param(
    default: tp.Any = dataclasses.MISSING,
    *,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
) -> tp.Any:
    return ref(
        "params",
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


@tp.overload
def dataclass(cls: tp.Type[A]) -> tp.Type[A]:
    ...


@tp.overload
def dataclass(
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
) -> tp.Callable[[tp.Type[A]], tp.Type[A]]:
    ...


@tpe.dataclass_transform(field_specifiers=(ref, param, field, static_field))
def dataclass(
    cls: tp.Optional[tp.Type[A]] = None,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
) -> tp.Union[tp.Type[A], tp.Callable[[tp.Type[A]], tp.Type[A]]]:
    decorator = dataclasses.dataclass(
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
    )

    if cls is None:
        return decorator

    return decorator(cls)



import dataclasses
import jax
from simple_pytree import Pytree
from nnx import partitioning
from nnx import scope_lib
import refx
import typing as tp
import jax.tree_util as jtu

M = tp.TypeVar("M", bound="Module")


class _ProxyContext(tp.Protocol):
    def __call__(self, __fn: tp.Callable[..., tp.Any], *args, **kwargs) -> tp.Any:
        ...


@dataclasses.dataclass
class CallableProxy:
    _proxy_context: _ProxyContext
    _proxy_callable: tp.Callable[..., tp.Any]

    def __call__(self, *args, **kwargs):
        return self._proxy_context(self._proxy_callable, *args, **kwargs)

    def __getattr__(self, name) -> "CallableProxy":
        return CallableProxy(self._proxy_context, getattr(self._proxy_callable, name))


class Module(Pytree):
    def deref(self: M) -> tp.Tuple[M, refx.DagDef]:
        return refx.deref(self)

    def reref(self: M, dagdef: refx.DagDef) -> M:
        return refx.reref(self, dagdef)

    def clone(self: M) -> M:
        return refx.clone(self)

    def make_rng(self, collection: str) -> jax.random.KeyArray:
        return scope_lib.make_rng(collection)

    def get_flag(self, name: str, default: tp.Any = dataclasses.MISSING) -> tp.Any:
        return scope_lib.get_flag(name, default)

    def __getitem__(self, filter: str) -> refx.Partition:
        return partitioning.get_partition(self.deref()[0], filter)

    def __setitem__(self, filter: str, value: refx.Partition):
        refx.update_refs(partitioning.get_partition(self, filter), value)

    def collections(self) -> tp.FrozenSet[str]:
        return frozenset(
            x.collection
            for x in jtu.tree_leaves(
                self, is_leaf=lambda x: isinstance(x, refx.Referential)
            )
            if isinstance(x, refx.Referential)
            if isinstance(x.collection, str)
        )

    @tp.overload
    def partition(
        self: M,
        *,
        is_leaf: tp.Optional[partitioning.LeafPredicate] = None,
    ) -> tp.Tuple[tp.Dict[str, refx.Partition], "ModuleDef[M]"]:
        ...

    @tp.overload
    def partition(
        self: M,
        collection: str,
        is_leaf: tp.Optional[partitioning.LeafPredicate] = None,
    ) -> tp.Tuple[refx.Partition, "ModuleDef[M]",]:
        ...

    @tp.overload
    def partition(
        self: M,
        collection: str,
        *extract_colelctions: str,
        is_leaf: tp.Optional[partitioning.LeafPredicate] = None,
    ) -> tp.Tuple[tp.Tuple[refx.Partition, ...], "ModuleDef[M]",]:
        ...

    def partition(
        self: M,
        *extract_collections: str,
        is_leaf: tp.Optional[partitioning.LeafPredicate] = None,
    ) -> tp.Tuple[
        tp.Union[
            tp.Dict[str, refx.Partition], tp.Tuple[refx.Partition, ...], refx.Partition
        ],
        "ModuleDef[M]",
    ]:
        collections = list(self.collections())
        module, dagdef = self.deref()
        partitions, treedef = partitioning.tree_partition(
            module, *collections, is_leaf=is_leaf
        )
        moduledef = ModuleDef(dagdef, treedef)

        if all(x is refx.NOTHING for x in partitions[-1].values()):
            partitions = partitions[:-1]
        else:
            # add "rest" as a reserved name for the rest of the tree
            collections.append("rest")

        partitions = dict(zip(collections, partitions))

        if extract_collections:
            if set(extract_collections) != set(collections):
                raise ValueError(
                    f"extract_colelctions contain all collections: "
                    f"{extract_collections} != {collections}"
                )

            if len(extract_collections) == 1:
                partitions = partitions[extract_collections[0]]
            else:
                partitions = tuple(partitions[x] for x in extract_collections)

        return partitions, moduledef

    @classmethod
    def init(
        cls: tp.Type[M],
        rngs: tp.Union[
            tp.Dict[str, jax.random.KeyArray], jax.random.KeyArray, None
        ] = None,
        **flags: tp.Hashable,
    ) -> tp.Type[M]:
        if rngs is None:
            rngs = {}
        elif isinstance(rngs, jax.Array):
            rngs = {"params": rngs}

        scope = scope_lib.Scope.from_keys_and_flags(rngs, flags)

        def _context(fn, *args, **kwargs):
            with scope_lib.scope(scope):
                return fn(*args, **kwargs)

        return CallableProxy(_context, cls)  # type: ignore

    def apply(
        self: M,
        *,
        rngs: tp.Optional[tp.Dict[str, jax.random.KeyArray]] = None,
        **flags: tp.Hashable,
    ) -> M:
        if rngs is None:
            rngs = {}

        scope = scope_lib.Scope.from_keys_and_flags(rngs, flags)

        def _context(fn, *args, **kwargs):
            with scope_lib.scope(scope):
                return fn(*args, **kwargs)

        return CallableProxy(_context, self)  # type: ignore


class ApplyCaller(tp.Protocol):
    def __getattr__(self, __name) -> "ApplyCaller":
        ...

    def __call__(
        self, *args, **kwargs
    ) -> tp.Tuple[tp.Any, tp.Dict[str, refx.Partition]]:
        ...


class ModuleDefValue(tp.NamedTuple):
    reref_dagdef: refx.DagDef
    partition_treedef: jtu.PyTreeDef


class ModuleDef(tp.Generic[M]):
    __slots__ = ("_reref_dagdef", "_partition_treedef")

    def __init__(self, reref_dagdef: refx.DagDef, partition_treedef: jtu.PyTreeDef):
        self._reref_dagdef = reref_dagdef
        self._partition_treedef = partition_treedef

    @property
    def reref_dagdef(self) -> refx.DagDef:
        return self._reref_dagdef

    @property
    def partition_treedef(self) -> jtu.PyTreeDef:
        return self._partition_treedef

    def apply(
        self,
        partitions: tp.Union[tp.Sequence[refx.Partition], tp.Dict[str, refx.Partition]],
        *,
        rngs: tp.Optional[tp.Dict[str, jax.random.KeyArray]] = None,
        flags: tp.Optional[tp.Dict[str, tp.Hashable]] = None,
    ) -> ApplyCaller:
        if flags is None:
            flags = {}
        if rngs is None:
            rngs = {}
        if isinstance(partitions, dict):
            partitions = tuple(partitions.values())
        module: M = self.merge(partitions)
        scope = scope_lib.Scope.from_keys_and_flags(rngs, flags)

        def _context(fn, *args, **kwargs):
            with scope_lib.scope(scope):
                out = fn(*args, **kwargs)
                partitions, _ = module.partition()
                return out, partitions

        return CallableProxy(_context, module)  # type: ignore

    def merge(
        self,
        partitions: tp.Union[tp.Sequence[refx.Partition], tp.Dict[str, refx.Partition]],
    ) -> M:
        if isinstance(partitions, dict):
            partitions = tuple(partitions.values())
        module: M = partitioning.merge_partitions(partitions, self.partition_treedef)
        return module.reref(self.reref_dagdef)


def _moduledef_flatten(
    x: ModuleDef[M],
) -> tp.Tuple[tp.Tuple[()], tp.Tuple[refx.DagDef, jtu.PyTreeDef]]:
    return (), (x.reref_dagdef, x.partition_treedef)


def _moduledef_unflatten(
    aux_data: tp.Tuple[refx.DagDef, jtu.PyTreeDef], _: tp.Tuple[()]
) -> ModuleDef[M]:
    return ModuleDef(*aux_data)


jtu.register_pytree_node(ModuleDef, _moduledef_flatten, _moduledef_unflatten)
